#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <deck.h>

//Uncomment this to show what is happening during execution:
//#define DETAILS
//Maximum size of matrices:
#define MAX 1000
//Maximum number of nodes:
#define MAX_NODES 16
//Message Size to use:
#define MSG_SIZE 8000

float a[MAX][MAX], b[MAX][MAX], c[MAX][MAX];
float buffer[MAX_NODES][((MAX*MAX)/sizeof(float))/2];
int n,nprocs;
deck_msg_t msg;
deck_mbox_t mbox;

void readdata()
{
	int i,j;
	FILE *fpa,*fpb;

	/* read matrix a  */
	fpa = fopen("a","r");
	for (i=0; i < n; i++) 
		for (j=0; j < n; j++) 
			fscanf(fpa,"%f", &a[i][j]);
	fclose(fpa);
	
	/* read matrix b */
	fpb = fopen("b","r");
	for (i=0; i < n; i++) 
		for (j=0; j < n; j++) 
			fscanf(fpb,"%f", &b[i][j]);
	fclose(fpb);
}

void printdata()
{
	int i,j;
	FILE *fpc;
	
	/* write results */
	fpc = fopen("deck_c","w");
	for (i=0; i < n; i++) {
		for (j = 0; j < n; j++) 
			fprintf(fpc,"%f\n",c[i][j]);
	}
	fclose(fpc);
}

void matmul()
{
	int i, j, k, m, id, my_id;
	int iprocs, jprocs;
	int i_id,j_id,ilb,iub,jlb,jub;
	int turn[MAX_NODES], nturns;
	
	for(i=0; i<MAX_NODES; i++)
		turn[i]=0;

	my_id = deck_node();
	/* number of processors in i direction */
	iprocs = (int) sqrt((double) nprocs);
	/* number of processors in j direction */
	jprocs = nprocs / iprocs;
	/* get processor ID in i and j dimensions */
	i_id = my_id % iprocs;
	j_id = my_id / iprocs; 
	/* find lower and upper bounds of i loop */
	ilb = i_id * n / iprocs;
	iub = (i_id + 1) * (n / iprocs);
	/* find lower and upper bounds of j loop */ 
	jlb = j_id * n / jprocs;
	jub = (j_id + 1) * (n / jprocs);

	#ifdef DETAILS
	printf("(node %i)parameters calculated.\nilb=%i\niub=%i\njlb=%i\njub=%i\n", deck_node(), ilb, iub, jlb, jub);
	fflush(stdout);
	#endif

	/* calculates the elements */
	for (i = ilb; i < iub; i++) {
		for (j = jlb; j < jub;  j++) {
			c[i][j] = 0.0;
			for (k = 0 ; k < n; k++) {
				c[i][j] += a[i][k] * b [k][j];
			}
		}
	}
	
	/* how many messages each node will send */
	if (nprocs > 0){
		if ((n*n/nprocs)%MSG_SIZE != 0) {
			nturns =  ((n*n/nprocs)/MSG_SIZE)+1;
		} else {
			nturns = (n*n/nprocs)/MSG_SIZE;
		}
	}
	if (deck_node() != 0){
		for (m = 0; m < (nprocs * nturns); m++){
		    /* synchronizes with other nodes */
		    if ((int)(m % nprocs) != 0){
			deck_barrier();
		    }
		    /* verify if it is my turn to send a message */
		    if (my_id == m % nprocs) {
			k = m / nprocs;

			#ifdef DETAILS
			printf("(node %i)m=%i,nturns=%i,nprocs=%i,packig msg %i\n", deck_node(), m, nturns, nprocs, k);
			fflush(stdout);
			#endif
			
			deck_msg_clear(&msg);
			deck_msg_pack(&msg, DECK_LONG, &my_id, 1);
			for (i = k * MSG_SIZE ; (i < (k+1)*MSG_SIZE) && (i < (n * n / nprocs)) ; i++) {
				deck_msg_pack(&msg, DECK_FLOAT, &c[ilb+(i/(jub-jlb))][jlb+i%(jub-jlb)], 1);
			}
			
			#ifdef DETAILS
			printf("(node %i)ready to send msg...\n", deck_node());
			fflush(stdout);
			#endif
			
			deck_mbox_post(&mbox, &msg);
			
			#ifdef DETAILS
			printf("(node %i)sent %i msg!\n", deck_node(), k);
			fflush(stdout);
			#endif
		    }
		}
	}else{
		/* so I am node 0 */
		/* let's receive messages of other nodes */
		for (k = 0 ; k < nturns*(nprocs-1) ; k++){
			deck_barrier();
			
			#ifdef DETAILS
			printf("(node %i)ready to retrv msg...\n", deck_node());
			fflush(stdout);
			#endif
			
			deck_mbox_retrv(&mbox, &msg);
			deck_msg_unpack(&msg, DECK_LONG, &id, 1);
			
			#ifdef DETAILS
			printf("(node %i)msg retrv'ed from node %i turn=%i\n", deck_node(), id, turn[id]);
			fflush(stdout);
			#endif
				
			/* put message on a buffer */
			deck_msg_unpack(&msg, DECK_FLOAT, &buffer[id][turn[id]*MSG_SIZE], MSG_SIZE);
			
			#ifdef DETAILS
			printf("(node %i)unpacked msg.\n", deck_node());
			fflush(stdout);
			#endif
			
			turn[id]++;
		}
		
		#ifdef DETAILS
		printf("(node %i)retrv'ed all msgs!\n", deck_node());
		for (i=1;i<nprocs;i++)
		fflush(stdout);
		#endif

		/* let's mount the result matrix */
		/* put other nodes results on our matix */
		for (id = 1; id < nprocs; id++) {
			// get processor ID in i and j dimensions 
			i_id = id % iprocs;
			j_id = id / iprocs; 
			// find lower and upper bounds of i loop 
			ilb = i_id * n / iprocs;
			iub = (i_id + 1) * (n / iprocs);
			// find lower and upper bounds of j loop  
			jlb = j_id * n / jprocs;
			jub = (j_id + 1) * (n / jprocs);
			k=0;
			for (i = ilb; i < iub; i++){
				for (j = jlb; j < jub; j++){
					c[i][j] = buffer[id][k];
					k++;
				}
			}
		}

		#ifdef DETAILS
		printf("(node %i)result matrix mounted.\n", deck_node());	
		fflush(stdout);
		#endif
		
	}
}

int main(int argc, char **argv)
{
	void readdata(),matmul(),printdata();

	deck_init(&argc, &argv);
	#ifdef DETAILS
	printf("(node %i)running...\n", deck_node());
	fflush(stdout);
	#endif
	/* get n and nprocs parameters */
	if (argc != 2){
		printf("usage: deck_matrix <size of matrices>\n");
		exit(1);
	}
	n = atoi(argv[1]);
	nprocs = deck_numnodes();
  
	/* creates DECK mailbox to receive messages */
	if (deck_node() == 0){
		deck_mbox_create(&mbox);
		deck_mbox_bind(&mbox, "node 0");
	} else {
			while (deck_mbox_fetch(&mbox, "node 0")	!= DECK_EOK)
				deck_thread_yield();
	}
	
	/* msg to send/receive results */
	deck_msg_create(&msg, MSG_SIZE*sizeof(float)+sizeof(long int), NULL);
	
	/* reads matrix a and b */
	readdata();
  
	#ifdef DETAILS
	printf("(node %i)matrices loaded...\n", deck_node());
	fflush(stdout);
	#endif
	
	/* do it! */
	matmul();
	
	/* outputs matrix c and destroys mailbox */
	if(deck_node() == 0){
		printdata();
		deck_mbox_destroy(&mbox);
	}
	
	#ifdef DETAILS
	printf("(node %i)done!\n", deck_node());
	fflush(stdout);
	#endif

	/* destroys message object */
	deck_msg_destroy(&msg);

	return (deck_done());
}

